#!/usr/bin/env python3

"""Configuration and context information for dragonglass"""

from argparse import Namespace
from pathlib import Path
from typing import Any

import tomllib

from .tree import SourceIndex, SourceNode


"""The default configuration file name."""
DEFAULT_CONFIG_FILE = ".dragonglass"
"""The default template directory name."""
DEFAULT_TEMPLATE_DIRECTORY = ".dragonglass.tmpl"
"""The default template name."""
DEFAULT_TEMPLATE_NAME = "default.html"
""" The default stylesheet name."""
DEFAULT_STYLESHEET_NAME = "dragonglass.css"
"""The default callout icon."""
CALLOUT_DEFICON = 'pencil'
"""The default callout icons."""
CALLOUT_ICONS = {
    'abstract': 'clipboard-list',
    'attention': 'triangle-alert',
    'bug': 'bug',
    'caution': 'triangle-alert',
    'check': 'check',
    'cite': 'quote',
    'danger': 'zap',
    'done': 'check',
    'error': 'zap',
    'example': 'list',
    'fail': 'x',
    'failure': 'x',
    'faq': 'circle-help',
    'help': 'circle-help',
    'info': 'info',
    'missing': 'x',
    'question': 'circle-help',
    'quote': 'quote',
    'success': 'check',
    'summary': 'clipboard-list',
    'tip': 'flame',
    'tldr': 'clipboard-list',
    'todo': 'circle-check',
    'warning': 'triangle-alert'
}

class Context:
    """
    The configuration and context information for dragonglass.

    Attributes:
        source_dir (Path): The source directory (Obsidian vault) for the operation.
        template_dir (Path): The directory in which templates may be found.
        config (dict[str, Any]): The configuration read in from the default configuration file.
        src_index (SourceIndex): The index of all nodes, used to look up targets for internal links.
        current_node (SourceNode): The current node being processed.
    """
    def __init__(self) -> None:
        """Initialize the Context."""
        self.source_dir: Path | None = None
        self.template_dir: Path | None = None
        self.config: dict[str, Any] = {}
        self.src_index: SourceIndex | None = None
        self.current_node: SourceNode | None = None
        self._default_template_name: str | None = None

    def load_config(self, args: Namespace) -> None:
        """
        Loads the configuration for dragonglass.

        Args:
            args (Namespace): The command line arguments to dragonglass.
        """
        config_filename: str = str(args.config) if args.config else DEFAULT_CONFIG_FILE
        assert self.source_dir is not None
        config_path = self.source_dir / config_filename
        if config_path.exists() and config_path.is_file():
            with open(config_path, "rb") as f:
                self.config = tomllib.load(f)
        templates_section = self.config.get("templates", {})
        self.template_dir = self.source_dir / templates_section.get("directory", DEFAULT_TEMPLATE_DIRECTORY)
        self._default_template_name = templates_section.get("default", DEFAULT_TEMPLATE_NAME)

    @property
    def default_stylesheet(self) -> str:
        """Returns the default stylesheet name."""
        templates_section = self.config.get("templates", {})
        return templates_section.get("stylesheet", DEFAULT_STYLESHEET_NAME)

    @property
    def description_title(self) -> bool:
        """Returns ``True`` if we use the title as the default description, ``False`` if not."""
        metadata_section = self.config.get("metadata", {})
        return metadata_section.get("description-title", False)

    @property
    def extern_in_new_tab(self) -> bool:
        """Returns ``True`` if external links (full URLs) will e opened in a new tab, ``False`` if not."""
        links_section = self.config.get("links", {})
        return links_section.get("extern-new-tab", False)

    @property
    def foreign_in_new_tab(self) -> bool:
        """Returns ``True`` if non-Markdown local files will be opened in a new tab, ``False`` if not."""
        links_section = self.config.get("links", {})
        return links_section.get("foreign-new-tab", False)

    @property
    def relative_links(self) -> bool:
        """
        Returns ``True`` if we generate relative links for internal links, ``False`` if we generate absolute links.
        """
        links_section = self.config.get("links", {})
        return links_section.get("relative", False)

    @property
    def site_base(self) -> str | None:
        """Returns the configured site base URL."""
        generate_section = self.config.get("metadata", {})
        return generate_section.get("sitebase", None)

    @property
    def site_title(self) -> str | None:
        """Returns the configured site title."""
        generate_section = self.config.get("metadata", {})
        return generate_section.get("sitetitle", None)

    @property
    def url_prefix(self) -> str:
        """Returns the configured URL prefix for all URLs generated for internal links."""
        links_section = self.config.get("links", {})
        rc = links_section.get("prefix", "/")
        return rc if rc.endswith("/") else rc + '/'

    def get_template_name_for_node(self, node: SourceNode) -> str:
        """
        Returns the name of a template to be used to render the current node.

        Args:
            node (SourceNode): The current node being processed.

        Returns:
            str: TRhe name of the template to be used to render the node.
        """
        if not self.template_dir.is_dir():
            return DEFAULT_TEMPLATE_NAME
        return node.metadata.get("template", self._default_template_name)

    def get_classname(self, classname: str) -> str:
        """
        Return a classname configured in the configuration file.

        Args:
            classname (str): The key for the classname to get, which is also the default value.

        Returns:
            str: the configured classname, or the ``classname`` argument itself if nothing was configured.
        """
        classnames_section = self.config.get("classnames", {})
        return classnames_section.get(classname, classname)

    def get_callout_iconname(self, callout: str) -> str:
        """
        Return the name for a Lucide icon that represents a particular callout type.

        Args:
            callout (str): The type of the callout in question.

        Returns:
            str: The name of an icon, or a default icon name.
        """
        callout_icons = self.config.get('callout-icons', {})
        return callout_icons.get(callout, CALLOUT_ICONS.get(callout, callout_icons.get('_default', CALLOUT_DEFICON)))
