#!/usr/bin/env python3

"""The Markdown parser and its extensions."""

import hashlib
import logging
import re
import xml.etree.ElementTree as etree  # noqa: N813
from typing import Any
from urllib.parse import unquote as urlunquote
from urllib.parse import quote as urlquote
from urllib.parse import urlparse

import markdown
from markdown import Markdown
from markdown.blockprocessors import BlockProcessor, BlockParser, BlockQuoteProcessor
from markdown.extensions import Extension
from markdown.extensions.footnotes import (FootnoteExtension, FootnoteBlockProcessor, FootnoteInlineProcessor,
                                           FootnoteTreeprocessor, FootnotePostTreeprocessor, FootnotePostprocessor)
from markdown.inlinepatterns import InlineProcessor, SimpleTagInlineProcessor
from markdown.preprocessors import Preprocessor
from markdown.util import STX, ETX

from .config import Context

logger = logging.getLogger(__name__)

PRIO_BASE = 10000  # priority base for our extensions

# Patterns for ObsidianImages
OBSIMAGE_PATTERN = r'!\[\[(.*?)\]\]'
GENERICIMAGE_PATTERN = r'!\[(.*?)\]\((.*?)\)'

# Patterns for ObsidianLinks
OBSLINK_PATTERN = r'\[\[(.*?)\]\]'
GENERICLINK_PATTERN = r'\[(.*?)\]\((.*?)\)'

# Inline footnote reference pattern.
INLINE_FOOTNOTE_REF_PREFIX = STX + "erbosoft_fnref:"
INLINE_FOOTNOTE_REF_PATTERN = INLINE_FOOTNOTE_REF_PREFIX + "{}" + ETX

# Obsidian comment marker
COMMENT_MARKER = '%%'

# Tags pattern
OBSTAG_PATTERN = r'#([a-zA-Z0-9/_-]+)'


def is_proper_url(s: str) -> bool:
    """
    Checks to see if a string is a "proper" URL.

    Args:
        s (str): The string to be parsed.

    Returns:
        bool: ``True`` if the string is a proper URL, ``False`` if not.
    """
    parseout = urlparse(s)
    return True if parseout.scheme else False


def sanitize_reference(s: str) -> str:
    """
    Sanitizes an internal reference to a file by removing URL-quoted characters and any Markdown suffix.

    Args:
        s (str): Th reference to be sanitized.

    Returns:
        str: The sanitized reference.
    """
    rc = urlunquote(s)
    if rc.endswith('.md'):
        rc = rc[:-3]
    return rc


def make_hash_value(s: str) -> str:
    """
    Create a secure hash value for the passed-in string.

    Args:
        s (str): The string to be hashed.

    Returns:
        A secure hash value for the string, expressed as hex digits.
    """
    return hashlib.sha1(bytes(s, 'utf-8')).hexdigest()


class MetaStripper(Extension):
    """
    An extension that strips the metadata off the front of Obsidian pages, as it's already been parsed in an
    earlier step.
    """

    class MetaStripperProc(Preprocessor):
        """The actual metadata stripper."""
        def run(self, lines: list[str]) -> list[str]:
            """
            Removes the metadata lines from the front of the array of lines.

            Args:
                lines (list[str]): The Markdown lies read in.

            Returns:
                list[str]: The Markdown lines after the metadata has been removed.
            """
            if lines[0] == '---':
                lines.pop(0)
                while lines[0] != '---':
                    lines.pop(0)
                lines.pop(0)
            return lines

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the metadata stripper with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the stripper with.
        """
        md.preprocessors.register(MetaStripper.MetaStripperProc(md), 'metastripper', PRIO_BASE)


class ObsidianHeaders(Extension):
    """An extension that augments the existing header processing."""

    class HashHeaderProcessor(BlockProcessor):
        """ Process Hash Headers. """

        # Detect a header at start of any line in block
        RE = re.compile(r'(?:^|\n)(?P<level>#{1,6})(?P<header>(?:\\.|[^\\])*?)#*(?:\n|$)')

        def test(self, parent: etree.Element, block: str) -> bool:
            """
            Tests to see whether the current block can be handled by this processor.

            Args:
                parent (etree.Element): The current parent element.
                block (str): The current block to be tested.

            Returns:
                bool: ``True`` if this processor can handle the block, ``False`` if not.
            """
            return bool(self.RE.search(block))

        def run(self, parent: etree.Element, blocks: list[str]) -> None:
            """
            Processes the text in the current block and adds elements to the parent element.

            Args:
                parent (etree.Element): Parent element to add new subelements to.
                blocks (list[str]): Blocks of text to be processed. The first block in this list is the one for which
                    the ``test`` method returned ``True``.  This method should remove any parsed text from the
                    ``block`` list before it returns.

            Returns:
                bool: ``True`` if text was parsed by this method, ``False`` if not.
            """
            block = blocks.pop(0)
            m = self.RE.search(block)
            if m:
                before = block[:m.start()]  # All lines before header
                after = block[m.end():]     # All lines after header
                if before:
                    # As the header was not the first line of the block and the
                    # lines before the header must be parsed first,
                    # recursively parse this lines as a block.
                    self.parser.parseBlocks(parent, [before])

                # Get the header text and its hash.
                hdr_text = m.group('header').strip()
                hval = make_hash_value(hdr_text)

                # Create header (with anchor) using named groups from RE
                a = etree.SubElement(parent, 'a', {'class': 'hdrref', 'name': f"hdr-{hval}"})
                h = etree.SubElement(a, 'h%d' % len(m.group('level')))
                h.text = hdr_text
                if after:
                    # Insert remaining lines as first block for future parsing.
                    if self.parser.state.isstate('looselist'):
                        # This is a weird edge case where a header is a child of a loose list
                        # and there is no blank line after the header. To ensure proper
                        # parsing, the line(s) after need to be detabbed. See #1443.
                        after = self.looseDetab(after)
                    blocks.insert(0, after)
            else:  # pragma: no cover
                # This should never happen, but just in case...
                logger.warn("We've got a problem header: %r" % block)

    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802
        """
        Registers the header processor with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the patterns with.
        """
        md.parser.blockprocessors.register(ObsidianHeaders.HashHeaderProcessor(md.parser), 'hashheader', 70)


class ObsidianImages(Extension):
    """An extension that supports image tags the way Obsidian handles them."""
    __DIMS = re.compile(r'(.*)\|[ ]*(\d+)(?:[ ]*x[ ]*(\d+))?')

    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        """
        Initialize the image processor class.

        Args:
            context (Context): The dragonglass context, containing the configuration.
            **kwargs (dict): Other keyword arguments to the image tag processor.
        """
        super(ObsidianImages, self).__init__(**kwargs)
        self._context = context

    @property
    def invalid_reference_classname(self) -> str:
        """
        Returns the CSS class name for an invalid reference in the text.
        """
        return self._context.get_classname('invalid-reference')

    def _parse_dimensions(self, s: str) -> tuple[str, int, int]:
        """
        Parses an image reference looking for image dimensions.

        Args:
            s (str): The image reference to be parsed.

        Returns:
            str: The actual image reference.
            int: The image width in pixels, or -1 if not specified.
            int: The image height in pixels, or -1 if not specified.
        """
        s = s.replace(r'\|', '|')  # handle case where we're inside tables
        m = self.__DIMS.match(s)
        if m:
            width = int(m.group(2))
            height = int(m.group(3)) if m.group(3) else -1
            return m.group(1), width, height
        else:
            return s, -1, -1

    def _lookup_image_reference(self, name: str) -> str | None:
        """
        Looks up the image reference in the source index and returns its link target value if it exists.

        Args:
            name (str): The image name to be looked up.

        Returns:
            str: The image's link target value, or ``None`` if the image was not found.
        """
        assert self._context.src_index is not None
        node, _ = self._context.src_index.lookup(self._context.current_node, name)
        if node:
            return node.link_target(self._context.url_prefix,
                                    self._context.current_node if self._context.relative_links else None)
        return None

    class ObsidianImageProc(InlineProcessor):
        """The actual inline processor for Obsidian image references."""
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            """
            Initialize the ObsidianImageProc processor.

            Args:
                pattern (str): The pattern to be matched to find the image reference.
                md (markdown.Markdown): Backreference to the Markdown processor.
                extref (ObsidianImages): External reference to the enclosing ObsidianImages object.
            """
            super(ObsidianImages.ObsidianImageProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            """
            Handles a match on the reference for this processor.

            Args:
                m (re.Match[str]): The regular expression match data.
                data (str): The entire block of text surrounding the pattern, as a multi-line string.

            Returns:
                el (etree.Element): The new HTML element being added to the tree, or ``None`` if the match was rejected.
                int: The index of the first character in ``data`` that was "consumed" by the pattern, or ``None``
                    if the match was rejected,
                int: The index of the first character in ``data`` that was *not* consumed by the pattern, or ``None``
                    if the match was rejected.
            """
            name, width, height = self._extref._parse_dimensions(m.group(1))
            link = self._extref._lookup_image_reference(name)
            if link is None:
                el = etree.Element('span')
                el.set('class', self._extref.invalid_reference_classname)
                el.text = name
            else:
                el = etree.Element('img')
                el.set('src', link)
                el.set('alt', name)
                if width > 0:
                    el.set('width', str(width))
                if height > 0:
                    el.set('height', str(height))
            return el, m.start(0), m.end(0)

    class GenericImageProc(InlineProcessor):
        """The actual inline processor for generic image references."""
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            """
            Initialize the GenericImageProc processor.

            Args:
                pattern (str): The pattern to be matched to find the image reference.
                md (markdown.Markdown): Backreference to the Markdown processor.
                extref (ObsidianImages): External reference to the enclosing ObsidianImages object.
            """
            super(ObsidianImages.GenericImageProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            """
            Handles a match on the reference for this processor.

            Args:
                m (re.Match[str]): The regular expression match data.
                data (str): The entire block of text surrounding the pattern, as a multi-line string.

            Returns:
                el (etree.Element): The new HTML element being added to the tree, or ``None`` if the match was rejected.
                int: The index of the first character in ``data`` that was "consumed" by the pattern, or ``None``
                    if the match was rejected,
                int: The index of the first character in ``data`` that was *not* consumed by the pattern, or ``None``
                    if the match was rejected.
            """
            name, width, height = self._extref._parse_dimensions(m.group(1))
            link = m.group(2)
            if is_proper_url(link):
                el = etree.Element('img')
                el.set('src', link)
                if len(name) > 0:
                    el.set('alt', name)
                if width > 0:
                    el.set('width', str(width))
                if height > 0:
                    el.set('height', str(height))
            else:
                newlink = self._extref._lookup_image_reference(link)
                if newlink:
                    el = etree.Element('img')
                    el.set('src', newlink)
                    if len(name) > 0:
                        el.set('alt', name)
                    if width > 0:
                        el.set('width', str(width))
                    if height > 0:
                        el.set('height', str(height))
                else:
                    el = etree.Element('span')
                    el.set('class', self._extref.invalid_reference_classname)
                    el.text = link
            return el, m.start(0), m.end(0)

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the image processor with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the footnote processor with.
        """
        md.inlinePatterns.register(ObsidianImages.ObsidianImageProc(OBSIMAGE_PATTERN, md, self),
                                   'obsidian_images', PRIO_BASE + 1010)
        md.inlinePatterns.register(ObsidianImages.GenericImageProc(GENERICIMAGE_PATTERN, md, self),
                                   'obsidian_generic_images', PRIO_BASE + 1000)


class ObsidianLinks(Extension):
    """
    An extension that processes Obsidian internal links in the [[page name]] format, as well as overrides the standard
    Markdown link processing to handle Obsidian internal links as well as external links.
    """
    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        """
        Initialize the ObsidianLinks extension.

        Args:
            context (Context): Context object that contains configuration information.
            **kwargs (dict[str, Any]: Additional configuration information.
        """
        super(ObsidianLinks, self).__init__(**kwargs)
        self._context = context

    @property
    def obsidian_link_classname(self) -> str:
        """Returns the classname for Obsidian links."""
        return self._context.get_classname('obsidian-link')

    @property
    def invalid_reference_classname(self) -> str:
        """Returns the classname for invalid references."""
        return self._context.get_classname('invalid-reference')

    @property
    def extern_link_target(self) -> str:
        """Returns the target attribute for external links, ot ``None`` if there is none."""
        return '_blank' if self._context.extern_in_new_tab else None

    def _parse_reference(self, contents: str) -> tuple[str | None, str, str | None]:
        """
        Parse a reference and break it into link target and title.

        Args:
            contents (str): Contents of a link reference to look up.

        Returns:
            str: The link target, or ``None`` if the link is invalid.
            str: The link title to be used.
            str: The target attribute for the resulting anchor, or ``None`` if not specified.
        """
        contents = contents.replace(r'\|','|')  # handle case where we're inside tables
        text = None
        t = contents.split('|')
        if len(t) > 1:
            text = t[1]
            contents = t[0].rstrip('\\')
        t = contents.split('#')
        hashloc = None
        if len(t) > 1:
            hashloc = make_hash_value(t[1].strip())
            contents = t[0]

        assert self._context.src_index is not None
        node, _ = self._context.src_index.lookup(self._context.current_node, contents)
        if not text:
            text = contents
        if node:
            # record a backlink from the current node to the found one
            node.backlinks.add(self._context.current_node)
            link = node.link_target(self._context.url_prefix,
                                    self._context.current_node if self._context.relative_links else None)
            if hashloc:
                link = f"{link}#hdr-{hashloc}"
            target_window = None
            if not node.is_md and self._context.foreign_in_new_tab:
                target_window = '_blank'
            return link, text, target_window
        return None, text, None

    class ObsidianLinksProc(InlineProcessor):
        """Processor that handles Obsidian links, [[page]]."""
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            """
            Initialize the ObsidianLinksProc.

            Args:
                pattern (str): Regular expression pattern to be matched by this processor.
                md (markdown.Markdown): Reference to the Markdown parser.
                extref (ObsidianLinks): Backreference to the outer object.
            """
            super(ObsidianLinks.ObsidianLinksProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            """
            Handles a match on the reference for this processor.

            Args:
                m (re.Match[str]): The regular expression match data.
                data (str): The entire block of text surrounding the pattern, as a multi-line string.

            Returns:
                el (etree.Element): The new HTML element being added to the tree, or ``None`` if the match was rejected.
                int: The index of the first character in ``data`` that was "consumed" by the pattern, or ``None``
                    if the match was rejected,
                int: The index of the first character in ``data`` that was *not* consumed by the pattern, or ``None``
                    if the match was rejected.
            """
            link, text, target = self._extref._parse_reference(m.group(1))
            if link is None:
                el = etree.Element('span')
                el.set('class', self._extref.invalid_reference_classname)
                el.text = text
            else:
                el = etree.Element('a')
                el.set('href', link)
                el.set('class', self._extref.obsidian_link_classname)
                if target:
                    el.set('target', target)
                el.text = text
            return el, m.start(0), m.end(0)

    class GenericLinksProc(InlineProcessor):
        """Processor that handles generic links, [link text](url)."""
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            """
            Initialize the GenericLinksProc.

            Args:
                pattern (str): Regular expression pattern to be matched by this processor.
                md (markdown.Markdown): Reference to the Markdown parser.
                extref (ObsidianLinks): Backreference to the outer object.
            """
            super(ObsidianLinks.GenericLinksProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            """
            Handles a match on the reference for this processor.

            Args:
                m (re.Match[str]): The regular expression match data.
                data (str): The entire block of text surrounding the pattern, as a multi-line string.

            Returns:
                el (etree.Element): The new HTML element being added to the tree, or ``None`` if the match was rejected.
                int: The index of the first character in ``data`` that was "consumed" by the pattern, or ``None``
                    if the match was rejected,
                int: The index of the first character in ``data`` that was *not* consumed by the pattern, or ``None``
                    if the match was rejected.
            """
            text = m.group(1)
            link = m.group(2)
            if link.startswith('<') and link.endswith('>'):  # handle whitespace encoding
                link = urlquote(link[1:-1].strip(), safe=':/')
            if is_proper_url(link):
                el = etree.Element('a')
                el.set('href', link)
                target = self._extref.extern_link_target
                if target:
                    el.set('target', target)
                el.text = text
            else:
                newlink, _, target = self._extref._parse_reference(sanitize_reference(link))
                if newlink is None:
                    el = etree.Element('span')
                    el.set('class', self._extref.invalid_reference_classname)
                    el.text = text
                else:
                    el = etree.Element('a')
                    el.set('href', newlink)
                    el.set('class', self._extref.obsidian_link_classname)
                    if target:
                        el.set('target', target)
                    el.text = text
            return el, m.start(0), m.end(0)

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the link inline processors with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the patterns with.
        """
        md.inlinePatterns.register(ObsidianLinks.ObsidianLinksProc(OBSLINK_PATTERN, md, self),
                                   'obsidian_links', PRIO_BASE + 110)
        md.inlinePatterns.register(ObsidianLinks.GenericLinksProc(GENERICLINK_PATTERN, md, self),
                                   'obsidian_generic_links', PRIO_BASE + 100)


class ObsidianLists(Extension):
    """An extension to process Obsidian lists, including making nested lists and task lists."""
    UL_RE = re.compile(r'^([ ]*)[-+*][ ]+(.*)')
    OL_RE = re.compile(r'^([ ]*)([1-9][0-9]*)\.[ ]+(.*)')
    TASK_RE = re.compile(r'^([ ]*)-[ ]\[(.)\][ ]+(.*)')
    LIST_START = STX + "erbosoft:lstart" + ETX
    LIST_END = STX + "erbosoft:lend" + ETX
    HORZ_RULES = ['***', '* * *', '---', '- - -', '___', '_ _ _']

    def is_horizontal_rule(self, line):
        """
        Returns ``True`` if this line is a horizontal rule, ``False`` if not.

        Args:
            line The line to be tested:
        """
        return line.lstrip() in self.HORZ_RULES

    def _find_listhead(self, line) -> tuple[str, int, int]:
        """
        Find if the specified line is a list element head.

        Args:
            line (str): The line to be checked.

        Returns:
            str: The list type, "ol" or "ul", or "" if this line is not a list head.
            int: The indent level of the list, or -1 if this line is not a list head.
            int: The start index of the list, or -1 if this line is not a list head for an ordered list.
        """
        m = self.UL_RE.match(line)
        if m:
            if self.is_horizontal_rule(line):
                # this is actually a horizontal rule, don't trip on it
                return '', -1, -1
            return 'ul', len(m.group(1)), -1
        m = self.OL_RE.match(line)
        if m:
            return 'ol', len(m.group(1)), int(m.group(2))
        return '', -1, -1

    class ObsidianListFinder(Preprocessor):
        """Preprocessor that finds and marks blocks in the text that contain lists."""
        def __init__(self, extref: Any) -> None:
            """
            Initialize the ObsidianListFinder.

            Args:
                extref (ObsidianLists): Reference to the outer object.
            """
            super(ObsidianLists.ObsidianListFinder, self).__init__()
            self._extref = extref

        def run(self, lines: list[str]) -> list[str]:
            """
            Execute the preprocessor on the Markdown text.

            Args:
                lines (list[str]): List of lines of text in the input document.

            Returns:
                list[str]: The list of lines to carry forward with.
            """
            i = 0
            in_list = False
            while i < len(lines):
                if in_list:
                    if (len(lines[i].strip()) == 0 or lines[i].startswith('#')
                            or self._extref.is_horizontal_rule(lines[i])):
                        in_list = False
                        lines.insert(i, self._extref.LIST_END)
                        i += 1
                else:
                    listtype, _, _ = self._extref._find_listhead(lines[i])
                    if len(listtype) > 0:
                        in_list = True
                        lines.insert(i, self._extref.LIST_START)
                        i += 1
                i += 1
            return lines

    class ObsidianListBlock(BlockProcessor):
        """The actual block processor that generates lists."""
        def __init__(self, parser: BlockParser, extref: Any) -> None:
            """
            Initialize the ObsidianListBlock.

            Args:
                parser (BlockParser): Reference to the block parser.
                extref (ObsidianLists): Backreference to the outer object.
            """
            super(ObsidianLists.ObsidianListBlock, self).__init__(parser)
            self._extref = extref

        @staticmethod
        def _attach_checkbox(listitem: etree.Element, data_line: int, cbox: str, text: str) -> None:
            """
            Attach a checkbox and its associated text to the ``li`` element.

            Args:
                listitem (etree.Element): The ``li`` element that's being worked on.
                data_line (int): Data line index for this list item.
                cbox (str): The checkbox marker character.
                text (str): The text to include after the checkbox.
            """
            checkbox = etree.SubElement(listitem, 'input')
            checkbox.attrib['type'] = 'checkbox'
            checkbox.attrib['class'] = 'task-list-item-checkbox'
            checkbox.attrib['aria-label'] = 'Task'
            checkbox.attrib['data-line'] = str(data_line)
            checkbox.attrib['data-task'] = cbox
            if not cbox.isspace():
                checkbox.attrib['checked'] = 'checked'
            if cbox == 'x':
                span = etree.SubElement(listitem, 'span')
                span.attrib['class'] = 'task-list-item-checked'
                span.text = text
            else:
                checkbox.tail = text

        def _build_element(self, parent: etree.Element, listtype: str, indent: int, data_line: int, lines: list[str]) -> None:
            """
            Builds a list element and adds it to the specified parent list.

            Args:
                parent (etree.element): The parent list element to insert "li" items under.
                listtype (str): The list type, either "ol" or "ul".
                indent (int): The indent level of the list, which is greater than or equal to 0.
                data_line (int): The data line index of the element.
                lines (list[str]): The lines comprising the list element.
            """
            textdata: list[str] = []
            cbox: str | None = None

            # Parse the header line to get the first bit of text.
            if listtype == 'ol':
                m = self._extref.OL_RE.match(lines[0])
                assert len(m.group(1)) == indent
                textdata.append(m.group(3))
            elif listtype == 'ul':
                m = self._extref.TASK_RE.match(lines[0])
                if m:
                    assert len(m.group(1)) == indent
                    cbox = m.group(2)
                    textdata.append(m.group(3))
                else:
                    m = self._extref.UL_RE.match(lines[0])
                    assert len(m.group(1)) == indent
                    textdata.append(m.group(2))

            # Build the list element
            listelement = etree.SubElement(parent, 'li')
            listelement.attrib['data-line'] = str(data_line)
            if cbox is not None:
                classname = 'task-list-item'
                if not cbox.isspace():
                    classname = f"{classname} is-checked"
                listelement.attrib['class'] = classname
            my_lines = list(lines)
            i = 1
            last_sublist: etree.Element | None = None
            while i < len(my_lines):
                subtype, subindent, _ = self._extref._find_listhead(my_lines[i])
                if len(subtype) == 0:
                    # ordinary text line - append it to the text data
                    textdata.append(my_lines[i])
                    i += 1
                else:
                    # start of new sublist
                    assert subindent > indent
                    # flush text data
                    if last_sublist is not None:
                        last_sublist.tail = '\n'.join(textdata)
                    elif cbox is not None:
                        self._attach_checkbox(listelement, data_line, cbox, '\n'.join(textdata))
                    else:
                        listelement.text = '\n'.join(textdata)
                    textdata = []
                    # parse the sublist and reset to parse the "excess"
                    last_sublist, my_lines = self._build_list(listelement, my_lines[i:])
                    i = 0

            # flush the text data one more time
            if len(textdata) > 0:
                if last_sublist is not None:
                    last_sublist.tail = '\n'.join(textdata)
                elif cbox is not None:
                    self._attach_checkbox(listelement, data_line, cbox, '\n'.join(textdata))
                else:
                    listelement.text = '\n'.join(textdata)

        def _build_list(self, parent: etree.Element, lines: list[str]) -> tuple[etree.Element, list[str]]:
            """
            Build a new list attached to the specified parent element.

            Args:
                parent (etree.Element): The parent element to build the list under.
                lines (list[str]): The lines of text comprising the list.

            Returns:
                etree.Element: The new list element.
                list[str]: The excess lines that can't be parsed as part of this list.
            """

            # Establish the start of the list.
            listtype, indent_level, start_index = self._extref._find_listhead(lines[0])
            assert indent_level >= 0
            list_top = etree.SubElement(parent, listtype)
            if listtype == 'ol' and start_index > 1:
                list_top.attrib['start'] = str(start_index)
            if listtype == 'ul':
                m = self._extref.TASK_RE.match(lines[0])
                if m:
                    list_top.attrib['class'] = 'contains-task-list'

            # The start index is set to -1 so, the first time around, we'll pick it up and set the start point.
            st = -1
            data_line = 0
            for i, line in enumerate(lines):
                new_type, new_indent, _ = self._extref._find_listhead(lines[i])
                if len(new_type) > 0:
                    if new_indent < indent_level:
                        # if the list head is lesser indented, this is the end of this element and the list
                        if st >= 0:
                            self._build_element(list_top, listtype, indent_level, data_line, lines[st:i])
                            data_line += 1
                        return list_top, lines[i:]
                    if new_indent == indent_level:
                        # this is the end of this element
                        if st >= 0:
                            self._build_element(list_top, listtype, indent_level, data_line, lines[st:i])
                            data_line += 1
                        if new_type != listtype:
                            # this is also the end of the list
                            return list_top, lines[i:]
                        st = i  # start of next element
            if st >= 0:  # end the final element
                self._build_element(list_top, listtype, indent_level, data_line, lines[st:])
            return list_top, []

        def test(self, parent: etree.Element, block: str) -> bool:
            """
            Tests to see whether the current block can be handled by this processor.

            Args:
                parent (etree.Element): The current parent element.
                block (str): The current block to be tested.

            Returns:
                bool: ``True`` if this processor can handle the block, ``False`` if not.
            """
            if self._extref.LIST_START in block:
                return True
            return False

        def run(self, parent: etree.Element, blocks: list[str]) -> bool:
            """
            Processes the text in the current block and adds elements to the parent element.

            Args:
                parent (etree.Element): Parent element to add new subelements to.
                blocks (list[str]): Blocks of text to be processed. The first block in this list is the one for which
                    the ``test`` method returned ``True``.  This method should remove any parsed text from the
                    ``block`` list before it returns.

            Returns:
                bool: ``True`` if text was parsed by this method, ``False`` if not.
            """
            chunk = blocks.pop(0)
            p = chunk.find(self._extref.LIST_START)
            self.parser.parseChunk(parent, chunk[:p])
            chunk = chunk[p + len(self._extref.LIST_START) + 1:]
            p = chunk.find(self._extref.LIST_END)
            while p < 0:
                chunk += '\n'
                chunk += blocks.pop(0)
                p = chunk.find(self._extref.LIST_END)
            blocks.insert(0, chunk[p + len(self._extref.LIST_END) + 1:])
            list_lines = chunk[:p].rstrip().split('\n')
            assert len(list_lines) > 0
            _, excess = self._build_list(parent, list_lines)
            if len(excess) > 0:
                blocks.insert(0, '\n'.join(excess))
            return True

    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802
        """
        Registers the list block processor with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the patterns with.
        """
        md.preprocessors.register(ObsidianLists.ObsidianListFinder(self), 'obsidian-list-finder', 18)
        md.parser.blockprocessors.register(ObsidianLists.ObsidianListBlock(md.parser, self),
                                           'obsidian-lists', PRIO_BASE + 50)


class ObsidianInlines(Extension):
    """An extension that handles the special Obsidian markdown format sequences."""
    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the inline patterns with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the patterns with.
        """
        md.inlinePatterns.register(SimpleTagInlineProcessor(r'()~~(.*?)~~', 'del'), 'strikeout', PRIO_BASE)
        md.inlinePatterns.register(SimpleTagInlineProcessor(r'()\=\=(.*?)\=\=', 'ins'), 'highlight', PRIO_BASE + 1)


class ObsidianComments(Extension):
    """An extension that removes Obsidian-style comments."""
    class CommentStripper(Preprocessor):
        """The actual preprocessor that removes comments."""
        def run(self, lines: list[str]) -> list[str]:
            """
            Removes the comments from the array of lines.

            Args:
                lines (list[str]): The Markdown lines read in.

            Returns:
                list[str]: The Markdown lines after the comments have been removed.
            """
            newlines: list[str] = []
            in_comment = False
            defer_in_comment = False
            for line in lines:
                arr = line.split(COMMENT_MARKER)
                if in_comment:
                    arr.pop(0)
                    in_comment = False
                if len(arr) % 2 == 0:
                    defer_in_comment = True
                    if len(arr) > 0:
                        arr.pop()
                if not in_comment:
                    i = len(arr) - 2
                    while i >= 0:
                        arr.pop(i)
                        i -= 2
                    newlines.append(''.join(arr))
                    if defer_in_comment:
                        in_comment = True
                        defer_in_comment = False
            return newlines

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the comment remover with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the comment remover with.
        """
        md.preprocessors.register(ObsidianComments.CommentStripper(md), 'comments-multi', PRIO_BASE + 1)


class ObsidianStyleFootnotes(FootnoteExtension):
    """
    An extension of the standard Markdown processor footnote extension to support Obsidian-style inline footnotes
    in addition to standard ones.
    """
    class BlockProcessor(FootnoteBlockProcessor):
        """A block processor extended to handle Obsidian inline footnotes."""
        RE_INLINE = re.compile(r'\^\[(.+)]')

        def __init__(self, footnotes: FootnoteExtension) -> None:
            super(ObsidianStyleFootnotes.BlockProcessor, self).__init__(footnotes)
            self._inlines_nonce = 0

        def run(self, parent: etree.Element, blocks: list[str]) -> bool:
            """ Find, set, and remove footnote definitions. """
            block = blocks.pop(0)
            m_block = self.RE.search(block)
            m_inline = self.RE_INLINE.search(block)
            if m_block or m_inline:

                if m_block is not None and m_inline is not None:
                    if m_block.start() > m_inline.start():
                        m_block = None  # inline goes first

                if m_block:
                    id = m_block.group(1)
                    fn_blocks = [m_block.group(2)]

                    # Handle rest of block
                    therest = block[m_block.end():].lstrip('\n')
                    m2 = self.RE.search(therest)
                    if m2:
                        # Another footnote exists in the rest of this block.
                        # Any content before match is continuation of this footnote, which may be lazily indented.
                        before = therest[:m2.start()].rstrip('\n')
                        fn_blocks[0] = '\n'.join([fn_blocks[0], self.detab(before)]).lstrip('\n')
                        # Add back to blocks everything from beginning of match forward for next iteration.
                        blocks.insert(0, therest[m2.start():])
                    else:
                        # All remaining lines of block are continuation of this footnote, which may be lazily indented.
                        fn_blocks[0] = '\n'.join([fn_blocks[0], self.detab(therest)]).strip('\n')

                        # Check for child elements in remaining blocks.
                        fn_blocks.extend(self.detectTabbed(blocks))

                    footnote = "\n\n".join(fn_blocks)
                    self.footnotes.setFootnote(id, footnote.rstrip())

                    if block[:m_block.start()].strip():
                        # Add any content before match back to blocks as separate block
                        blocks.insert(0, block[:m_block.start()].rstrip('\n'))
                    return True

                if m_inline:
                    footnote = m_inline.group(1).strip()
                    id = make_hash_value(f"{footnote}{self._inlines_nonce}")
                    self._inlines_nonce += 1
                    self.footnotes.setFootnote(id, footnote)
                    blocks.insert(0, block[:m_inline.start(0)].rstrip()
                                  + INLINE_FOOTNOTE_REF_PATTERN.format(id) + block[m_inline.end(0):])
                    return True

            # No match. Restore block.
            blocks.insert(0, block)
            return False

    class InlineReplacer(InlineProcessor):
        def __init__(self, pattern: str, extref: Any) -> None:
            super(ObsidianStyleFootnotes.InlineReplacer, self).__init__(pattern)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            etxpoint = data.find(ETX, m.end(0))
            assert etxpoint > m.end(0)
            id = data[m.end():etxpoint]
            assert id in self._extref.footnotes.keys()
            sup = etree.Element('sup')
            a = etree.SubElement(sup, 'a')
            sup.set('id', self._extref.makeFootnoteRefId(id, found=True))
            a.set('href', '#' + self._extref.makeFootnoteId(id))
            a.set('class', 'footnote-ref')
            a.text = self._extref.getConfig('SUPERSCRIPT_TEXT').format(
                list(self._extref.footnotes.keys()).index(id) + 1
            )
            return sup, m.start(0), etxpoint + 1

    def extendMarkdown(self, md) -> None:
        """
        Registers the footnote processor with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the footnote processor with.
        """
        md.registerExtension(self)
        self.parser = md.parser
        self.md = md
        # Insert a `blockprocessor` before `ReferencePreprocessor`
        md.parser.blockprocessors.register(ObsidianStyleFootnotes.BlockProcessor(self), 'footnote', 17)

        # Insert an inline pattern before `ImageReferencePattern`
        FOOTNOTE_RE = r'\[\^([^\]]*)\]'  # blah blah [^1] blah
        md.inlinePatterns.register(FootnoteInlineProcessor(FOOTNOTE_RE, self), 'footnote', 175)
        md.inlinePatterns.register(ObsidianStyleFootnotes.InlineReplacer(INLINE_FOOTNOTE_REF_PREFIX, self),
                                   'footnote2', 175)
        # Insert a tree-processor that would actually add the footnote div
        # This must be before all other tree-processors (i.e., `inline` and
        # `codehilite`) so they can run on the the contents of the div.
        md.treeprocessors.register(FootnoteTreeprocessor(self), 'footnote', 50)

        # Insert a tree-processor that will run after inline is done.
        # In this tree-processor we want to check our duplicate footnote tracker
        # And add additional `backrefs` to the footnote pointing back to the
        # duplicated references.
        md.treeprocessors.register(FootnotePostTreeprocessor(self), 'footnote-duplicate', 15)

        # Insert a postprocessor after amp_substitute processor
        md.postprocessors.register(FootnotePostprocessor(self), 'footnote', 25)


class ObsidianStyleBlockquotes(Extension):
    """Extension that handles both blockquotes and callouts in the Obsidian style."""
    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        """
        Initialize the ObsidianStyleBlockquotes extension.

        Args:
            context (Context): Context object that contains configuration information.
            **kwargs (dict[str, Any]: Additional configuration information.
        """
        super(ObsidianStyleBlockquotes, self).__init__(**kwargs)
        self._context = context

    def _callout_iconname(self, callout: str) -> str:
        """
        Returns the Lucide icon name for a specific type of callout.

        Args:
            callout (str): The callout type.

        Returns:
            The icon name to use.
        """
        return self._context.get_callout_iconname(callout)

    class ObsidianBlockQuote(BlockQuoteProcessor):
        CALLOUT = re.compile(r'^\[!([a-z]+)\]([-+])?(?:[ ]+(.*))?')

        def __init__(self, parser: BlockParser, extref: Any) -> None:
            super(ObsidianStyleBlockquotes.ObsidianBlockQuote, self).__init__(parser)
            self._extref = extref

        def normal_blockquote(self, parent: etree.Element, block: str) -> None:
            sibling = self.lastChild(parent)
            if sibling is not None and sibling.tag == "blockquote":
                # Previous block was a blockquote so set that as this blocks parent
                quote = sibling
            else:
                # This is a new blockquote. Create a new parent element.
                quote = etree.SubElement(parent, 'blockquote')
            # Recursively parse block with blockquote as parent.
            # change parser state so blockquotes embedded in lists use `p` tags
            self.parser.state.set('blockquote')
            self.parser.parseChunk(quote, block)
            self.parser.state.reset()

        def callout_block(self, parent: etree.Element, lines: list[str]) -> None:
            m = self.CALLOUT.match(lines.pop(0))
            callout_type = m.group(1)
            folding = m.group(2)
            title = m.group(3)
            if not title:
                title = callout_type.title()
            baseclass = 'callout'
            if folding:
                baseclass = f"{baseclass} is-collapsible"
                if folding == '-':
                    baseclass = f"{baseclass} is-collapsed"
            base_div = etree.SubElement(parent, 'div', {'class': baseclass, 'data-callout': callout_type})
            title_div = etree.SubElement(base_div, 'div', {'class': 'callout-title'})
            icon_div = etree.SubElement(title_div, 'div', {'class': 'callout-icon'})
            etree.SubElement(icon_div, 'span',
                             {'data-lucide': self._extref._callout_iconname(callout_type)})
            inner_title_div = etree.SubElement(title_div, 'div', {'class': 'callout-title-inner'})
            inner_title_div.text = title
            if folding:
                baseclass = 'callout-fold'
                if folding == '-':
                    baseclass = f"{baseclass} is-collapsed"
                fold_div = etree.SubElement(title_div, 'div', {'class': baseclass})
                etree.SubElement(fold_div, 'span',
                                 {'data-lucide': 'chevron-right' if folding == '-' else 'chevron-down'})
            if len(lines) > 0:
                content_div = etree.SubElement(base_div, 'div', {'class': 'callout-content'})
                if folding:
                    styleval = 'display: none;' if folding == '-' else ''
                    content_div.attrib['style'] = styleval
                self.parser.state.set('callout')
                self.parser.parseBlocks(content_div, ['\n'.join(lines)])
                self.parser.state.reset()

        def run(self, parent: etree.Element, blocks: list[str]) -> None:
            block = blocks.pop(0)
            lines: list[str] = []
            callout = False
            m = self.RE.search(block)
            if m:
                before = block[:m.start()]  # Lines before blockquote
                # Pass lines before blockquote in recursively for parsing first.
                self.parser.parseBlocks(parent, [before])
                # Remove `> ` from beginning of each line.
                lines = [self.clean(line) for line in block[m.start():].split('\n')]
                callout = (self.CALLOUT.match(lines[0]) is not None)
            if callout:
                self.callout_block(parent, lines)
            else:
                self.normal_blockquote(parent, '\n'.join(lines))

    class CalloutLinesProcessor(BlockProcessor):
        def test(self, parent: etree.Element, block: str) -> bool:
            return self.parser.state.isstate('callout')

        def run(self, parent: etree.Element, blocks: list[str]) -> bool:
            block = blocks.pop(0)
            if block.strip():
                sibling = self.lastChild(parent)
                lines = block.split('\n')
                for line in lines:
                    if sibling is not None:
                        if sibling.tail:
                            sibling = etree.SubElement(parent, 'br')
                    else:
                        if parent.text:
                            sibling = etree.SubElement(parent, 'br')
                        else:
                            parent.text = line
                            continue
                    sibling.tail = line
            return True

    def extendMarkdown(self, md) -> None:
        md.parser.blockprocessors.register(ObsidianStyleBlockquotes.ObsidianBlockQuote(md.parser, self), 'quote', 20)
        md.parser.blockprocessors.register(ObsidianStyleBlockquotes.CalloutLinesProcessor(md.parser),
                                           'callout-text', 11)


class ObsidianTags(Extension):
    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        """
        Initialize the ObsidianStyleBlockquotes extension.

        Args:
            context (Context): Context object that contains configuration information.
            **kwargs (dict[str, Any]: Additional configuration information.
        """
        super(ObsidianTags, self).__init__(**kwargs)
        self._context = context

    def stash_tag(self, tagname: str) -> None:
        self._context.current_node.stash_tag(tagname)

    class ObsidianTagsProcessor(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            """
            Initialize the GenericLinksProc.

            Args:
                pattern (str): Regular expression pattern to be matched by this processor.
                md (markdown.Markdown): Reference to the Markdown parser.
                extref (ObsidianLinks): Backreference to the outer object.
            """
            super(ObsidianTags.ObsidianTagsProcessor, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element | None, int | None, int | None]:  # noqa: N802
            tagname = m.group(1)
            if not re.search(r'[^0-9]', tagname):
                return None, None, None
            self._extref.stash_tag(tagname)
            tag = etree.Element('span')
            tag.attrib['class'] = "tag"
            tag.text = f"#{tagname}"
            return tag, m.start(0), m.end(0)

    def extendMarkdown(self, md) -> None:
        md.inlinePatterns.register(ObsidianTags.ObsidianTagsProcessor(OBSTAG_PATTERN, md, self),
                                   'obsidian-tags', PRIO_BASE + 40)


def create_markdown_parser(context: Context) -> markdown.Markdown:
    """
    Creates a Markdown parser with all our extensions loaded.

    Args:
        context (Context): The dragonglass context, that contains the configuration.

    Returns:
        markdown.Markdown: The new Markdown parser.
    """
    extconfig: dict[str, Any] = {
        'codehilite': {
            'linenums': False,
            'guess_lang': False,
            'noclasses': False,
            'use_pygments': True,
            'lang_prefix': 'language'
        }
    }
    return markdown.Markdown(extensions=['fenced_code',
                                         'codehilite',
                                         'sane_lists',
                                         'tables',
                                         MetaStripper(),
                                         ObsidianComments(),
                                         ObsidianHeaders(),
                                         ObsidianStyleFootnotes(SUPERSCRIPT_TEXT='[{}]', SEPARATOR='-'),
                                         ObsidianStyleBlockquotes(context),
                                         ObsidianImages(context),
                                         ObsidianLinks(context),
                                         ObsidianLists(),
                                         ObsidianInlines(),
                                         ObsidianTags(context)],
                             extension_configs=extconfig)
